/*****************************************************************************
 *
 * sxInit.c -- Initialise Starlight Xpress filter wheels
 *
 * Copyright 2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>

#include <libusb-1.0/libusb.h>

#include <errno.h>
#include <openastro/filterwheel.h>

#include "sxfw.h"
#include "oafwprivate.h"
#include "unimplemented.h"

#define	wheelState		wheel->_sx


static void _sxInitFunctionPointers ( oaFilterWheel* );
static int  _getNumSlots ( oaFilterWheel* );

/**
 * Initialise a given filter wheel device
 */

oaFilterWheel*
oaSXInitFilterWheel ( oaFilterWheelDevice* device )
{
  int					i, j, k, l, res, matched;
  int					deviceAddr, deviceBus, numUSBDevices;
  int					interfaceNum, endpointOut;
  int					endpointIn, maxPacketSize;
  libusb_device**			devlist;
  libusb_device*			usbDevice;
  libusb_device_handle*			usbHandle = 0;
  struct libusb_device_descriptor	desc;
  oaFilterWheel*			wheel;

  if (!( wheel = ( oaFilterWheel* ) malloc ( sizeof ( oaFilterWheel )))) {
    perror ( "malloc oaFilterWheel failed" );
    return 0;
  }

  OA_CLEAR ( wheelState );
  OA_CLEAR ( wheel->features );


  pthread_mutex_init ( &wheelState.usbMutex, 0 );

  ( void ) strcpy ( wheel->deviceName, device->deviceName );
  ( void ) strcpy ( wheelState.devicePath, device->sysPath );
  wheelState.initialised = 0;
  wheelState.index = -1;

  libusb_init ( &wheelState.usbContext );
  // libusb_set_debug ( wheelState.usbContext, LIBUSB_LOG_LEVEL_DEBUG );
  numUSBDevices = libusb_get_device_list ( wheelState.usbContext, &devlist );
  if ( numUSBDevices < 1 ) {
    libusb_free_device_list ( devlist, 1 );
    libusb_exit ( wheelState.usbContext );
    if ( numUSBDevices ) {
      fprintf ( stderr, "Can't see any USB devices now (list returns -1)\n" );
      free ( wheel );
      return 0;
    }
    fprintf ( stderr, "Can't see any USB devices now\n" );
    free ( wheel );
    return 0;
  }

  matched = 0;
  deviceAddr = device->_devIndex & 0xff;
  deviceBus = device->_devIndex >> 8;
  for ( i = 0; i < numUSBDevices && !matched; i++ ) {
    struct libusb_config_descriptor *configDesc = 0;

    usbDevice = devlist[i];
    if ( LIBUSB_SUCCESS != libusb_get_device_descriptor ( usbDevice, &desc )) {
      libusb_free_device_list ( devlist, 1 );
      libusb_exit ( wheelState.usbContext );
      fprintf ( stderr, "get device descriptor failed\n" );
      free ( wheel );
      return 0;
    }
    if ( desc.idVendor == STARLIGHT_XPRESS_VID &&
        desc.idProduct == STARLIGHT_XPRESS_FILTERWHEEL_PID &&
        libusb_get_bus_number ( usbDevice ) == deviceBus &&
        libusb_get_device_address ( usbDevice ) == deviceAddr ) {
      // this looks like the one!

      if ( desc.bDeviceClass != LIBUSB_CLASS_PER_INTERFACE ) {
        fprintf ( stderr, "%s: device class is not expected type\n",
            __FUNCTION__ );
        free ( wheel );
        return 0;
      }

      if (( res = libusb_get_active_config_descriptor ( usbDevice,
          &configDesc )) < 0 ) {
        libusb_get_config_descriptor ( usbDevice, 0, &configDesc );
      }
      if ( configDesc ) {
        for ( j = 0; j < configDesc->bNumInterfaces; j++ ) {
          const struct libusb_interface *intf = &configDesc->interface[j];
          for ( k = 0; k < intf->num_altsetting; k++ ) {
            const struct libusb_interface_descriptor *intf_desc;
            intf_desc = &intf->altsetting[k];
            if ( intf_desc->bInterfaceClass == LIBUSB_CLASS_HID ) {
              interfaceNum = intf_desc->bInterfaceNumber;
              endpointIn = endpointOut = 0;
              for ( l = 0; !endpointIn && !endpointOut &&
                  l < intf_desc->bNumEndpoints; l++ ) {
                int attribs = intf_desc->endpoint[l].bmAttributes;
                int epAddress = intf_desc->endpoint[l].bEndpointAddress;
                int isInterrupt = ( attribs & LIBUSB_TRANSFER_TYPE_MASK )
                    == LIBUSB_TRANSFER_TYPE_INTERRUPT ? 1 : 0;
                int isOutput = ( epAddress & LIBUSB_ENDPOINT_DIR_MASK )
                    == LIBUSB_ENDPOINT_OUT ? 1 : 0;
                int isInput = ( epAddress & LIBUSB_ENDPOINT_DIR_MASK )
                    == LIBUSB_ENDPOINT_IN ? 1 : 0;
                if ( !endpointIn && isInterrupt && isInput ) {
                  endpointIn = epAddress;
                  maxPacketSize = intf_desc->endpoint[l].wMaxPacketSize;
                }
                if ( !endpointOut && isInterrupt && isOutput ) {
                  endpointOut = epAddress;
                }
              }
              matched = 1;
              libusb_open ( usbDevice, &usbHandle );
            }
          }
        }
        libusb_free_config_descriptor ( configDesc );
      }
    }
  }
  libusb_free_device_list ( devlist, 1 );
  if ( !matched ) {
    fprintf ( stderr, "No matching USB device found!\n" );
    libusb_exit ( wheelState.usbContext );
    free ( wheel );
    return 0;
  }
  if ( !usbHandle ) {
    fprintf ( stderr, "Unable to open USB device!\n" );
    libusb_exit ( wheelState.usbContext );
    free ( wheel );
    return 0;
  }

  if ( libusb_kernel_driver_active ( usbHandle, interfaceNum )) {
      libusb_detach_kernel_driver( usbHandle, interfaceNum );
  }
  if ( libusb_claim_interface ( usbHandle, interfaceNum )) {
    fprintf ( stderr, "Unable to claim interface for USB device!\n" );
    libusb_exit ( wheelState.usbContext );
    free ( wheel );
    return 0;
  }

  wheel->interface = device->interface;
  wheelState.usbHandle = usbHandle;
  wheelState.endpointIn = endpointIn;
  wheelState.maxPacketSize = maxPacketSize;
  wheelState.endpointOut = endpointOut;
  wheelState.index = device->_devIndex;
  wheelState.wheelType = device->_devType;
  wheelState.interfaceNum = interfaceNum;
  wheelState.currentPosition = 1;

  if (( wheelState.numSlots = _getNumSlots ( wheel )) < 1 ) {
    libusb_release_interface ( usbHandle, interfaceNum );
    fprintf ( stderr, "%s: invalid number of slots in filter wheel\n",
        __FUNCTION__ );
    return 0;
  }

  _oaInitFilterWheelFunctionPointers ( wheel );
  _sxInitFunctionPointers ( wheel );

  wheel->features.coldReset = 0;
  wheel->features.warmReset = 0;
  wheel->features.speedControl = 0;

  return wheel;
}


static void
_sxInitFunctionPointers ( oaFilterWheel* wheel )
{
  wheel->funcs.initWheel = oaSXInitFilterWheel;
  wheel->funcs.getName = oaSXWheelGetName;
  wheel->funcs.closeWheel = oaSXWheelClose;
  // wheel->funcs.moveForward = _moveForward;
  // wheel->funcs.moveBackward = _moveBackward;
  wheel->funcs.moveTo = oaSXMoveTo;
  // wheel->funcs.warmReset = oaSXWheelWarmReset;
  // wheel->funcs.coldReset = oaSXWheelColdReset;
  wheel->funcs.numSlots = oaSXNumSlots;
}


int
_getNumSlots ( oaFilterWheel* wheel )
{
  char buffer[2];
  int numRead, numSlots;

  pthread_mutex_lock ( &wheelState.usbMutex );

  buffer[0] = 0;
  buffer[1] = 1;
  if ( _sxWheelWrite ( wheel, buffer, 2 ) != 2 ) {
    fprintf ( stderr, "%s: write error on get filters command\n",
      __FUNCTION__ );
    pthread_mutex_unlock ( &wheelState.usbMutex );
    return 0;
  }

  numRead = _sxWheelRead ( wheel, buffer, 2 );
  if ( numRead != 2 ) {
    pthread_mutex_unlock ( &wheelState.usbMutex );
    return 0;
  }

  do {
    sleep ( 1 );
    buffer[0] = 0;
    buffer[1] = 0;
    if ( _sxWheelWrite ( wheel, buffer, 2 ) != 2 ) {
      fprintf ( stderr, "%s: write error on get filters command\n",
        __FUNCTION__ );
      pthread_mutex_unlock ( &wheelState.usbMutex );
      return 0;
    }

    numRead = _sxWheelRead ( wheel, buffer, 2 );
    if ( numRead != 2 ) {
      pthread_mutex_unlock ( &wheelState.usbMutex );
      return 0;
    }
  } while ( !buffer[0] );

  pthread_mutex_unlock ( &wheelState.usbMutex );

  numSlots = buffer[1];
  return numSlots;
}


int
oaSXWheelClose ( oaFilterWheel* wheel )
{
  wheelState.initialised = 0;
  libusb_release_interface ( wheelState.usbHandle, wheelState.interfaceNum );
  libusb_close ( wheelState.usbHandle );
  libusb_exit ( wheelState.usbContext );
  return 0;
}
