/*****************************************************************************
 *
 * filterwheel.cc -- filter wheel interface class
 *
 * Copyright 2014 James Fidell (james@openastroproject.org)
 *
 * License:
 *
 * This file is part of the Open Astro Project.
 *
 * The Open Astro Project is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The Open Astro Project is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the Open Astro Project.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *****************************************************************************/

#include <oa_common.h>

#include "filterWheel.h"
#include "configuration.h"
#include "state.h"
#include "version.h"


#define wheelFuncs     wheelContext->funcs

FilterWheel::FilterWheel()
{
  initialised = 0;
}


FilterWheel::~FilterWheel()
{
}


// FIX ME -- might be nice to make this a tidier type at some point.  vector?
int
FilterWheel::listConnected ( oaFilterWheelDevice** devs )
{
  return ( oaGetFilterWheels ( devs ));
}


int
FilterWheel::initialise ( oaFilterWheelDevice* device )
{
  disconnect();

  if ( !device ) {
    qWarning() << "device is null!";
    return -1;
  }

  if (( wheelContext = device->initFilterWheel ( device ))) {
    initialised = 1;
    // FIX ME -- this lot should probably be done in the caller
    if ( state.settingsWidget ) {
      int numSlots = wheelFuncs.numSlots ( wheelContext );
      state.settingsWidget->setSlotCount ( numSlots );
      state.captureWidget->setSlotCount ( numSlots );
    }
    if ( state.captureWidget ) {
      state.captureWidget->reloadFilters();
    }
    return 0;
  }
  return -1;
}


void
FilterWheel::disconnect ( void )
{
  if ( initialised ) {
    wheelFuncs.closeWheel ( wheelContext );
    initialised = 0;
    // FIX ME -- this lot should probably be done in the caller
    if ( state.settingsWidget ) {
      state.settingsWidget->setSlotCount ( 0 );
      state.captureWidget->setSlotCount ( 0 );
    }
    if ( state.captureWidget ) {
      state.captureWidget->reloadFilters();
    }
  }
}


const char*
FilterWheel::name ( void )
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with filter wheel uninitialised";
    return "";
  }
  return wheelFuncs.getName ( wheelContext );
}


int
FilterWheel::isInitialised ( void )
{
  return initialised;
}


void
FilterWheel::warmReset ( void )
{
  if ( initialised ) {
    wheelFuncs.warmReset ( wheelContext, 0 );
  }
}


void
FilterWheel::coldReset ( void )
{
  if ( initialised ) {
    wheelFuncs.coldReset ( wheelContext, 0 );
  }
}


int
FilterWheel::numSlots ( void )
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with filter wheel uninitialised";
    return -1;
  }
  return wheelFuncs.numSlots ( wheelContext );
}


int
FilterWheel::selectFilter ( int slotNum )
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with filter wheel uninitialised";
    return -1;
  }
  return wheelFuncs.moveTo ( wheelContext, slotNum, 0 );
}


int
FilterWheel::hasWarmReset ( void )
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with filter wheel uninitialised";
    return -1;
  }
  return wheelContext->features.warmReset;
}


int
FilterWheel::hasColdReset ( void )
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with filter wheel uninitialised";
    return -1;
  }
  return wheelContext->features.coldReset;
}


int
FilterWheel::hasSpeedControl ( void )
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with filter wheel uninitialised";
    return -1;
  }
  return wheelContext->features.speedControl;
}


int
FilterWheel::getSpeed ( unsigned int* speed )
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with filter wheel uninitialised";
    return -1;
  }

  return wheelFuncs.getSpeed ( wheelContext, speed );
}


int
FilterWheel::setSpeed ( unsigned int speed, int nodelay )
{
  if ( !initialised ) {
    qWarning() << __FUNCTION__ << " called with filter wheel uninitialised";
    return -1;
  }

  if ( speed > 100 ) { speed = 100; }
  return wheelFuncs.setSpeed ( wheelContext, speed, nodelay );
}


void
FilterWheel::updateSearchFilters ( int interfaceType )
{
  int numFilters;

  oaClearFilterWheelIDFilters ( interfaceType );
  numFilters = config.filterWheelConfig[ interfaceType ].count();
  if ( numFilters ) {
    for ( int i = 0; i < numFilters; i++ ) {
      oaAddFilterWheelIDFilter ( interfaceType,
          &( config.filterWheelConfig[ interfaceType ][ i ] ));
    }
  }
}


void
FilterWheel::updateAllSearchFilters ( void )
{
  for ( int i = 1; i < OA_FW_IF_COUNT; i++ ) {
    updateSearchFilters ( i );
  }
}
